//
//  CWPoint.m
//  ResaISkane
//
//  Copyright 2009-2010 Fredrik Olsson. All rights reserved.
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.
//

#import "CWPoint.h"
#import "CWXMLTranslater.h"
#import <CoreLocation/CoreLocation.h>
#import <MapKit/MapKit.h>
#import "NSString+CWLocalizedFormats.h"

NSString *const CWPointStatusDidChangeNotification = @"CWPointStatusChangesNotification";

@interface CWPointParser : NSObject <CWXMLTranslaterDelegate> {
@private
  NSURL* url;
}

-(id)initWithSearchString:(NSString*)searchString;

-(NSArray*)parsePoints;

@end

@interface CWPoint () <CLLocationManagerDelegate, MKReverseGeocoderDelegate>

@end


@implementation CWPoint

@synthesize ID = _ID;
@synthesize name = _name;
@synthesize stopPoint = _stopPoint;
@synthesize type = _type;

-(NSString*)typeString;
{
  switch (self.type) {
    case CWPointTypeStopArea:
      return @"STOP_AREA";
    case CWPointTypeAddress:
      return @"ADDRESS";
    case CWPointTypePointOfInterest:
      return @"POI";
    default:
      return @"_unknown_";
  }
}

// NOTE: Journey queries must report ADDRESS and POI as same type, for some stupid reason.
static int CWQueryTypeForPointType(CWPointType type) {
  if (type == 0) {
    return 0;
  } else {
    return 1;
  }
}

static CLLocation* lastLocation = nil;
static BOOL failedLocation = NO;
static CLLocationManager* locationManager = nil;
static BOOL isUpdatingLocation = NO;

- (void)locationManager:(CLLocationManager *)manager didFailWithError:(NSError *)error
{
	if ([error code] != kCLErrorLocationUnknown) {
		failedLocation = YES;
    if ([error code] != kCLErrorDenied) {
      UIAlertView* alert = [[UIAlertView alloc] initWithTitle:NSLocalizedString(@"LocationError", nil)
                                                      message:NSLocalizedString(@"LocationErrorMessage", nil)
                                                     delegate:nil
                                            cancelButtonTitle:NSLocalizedString(@"OK", nil)
                                            otherButtonTitles:nil];
      [alert show];
      [alert release];
    }
    [manager stopUpdatingLocation];
    [[NSNotificationCenter defaultCenter] postNotificationName:CWPointStatusDidChangeNotification object:[CWPoint currentLocationPoint]];
  }
}

static MKReverseGeocoder* geocoder = nil;

- (void)reverseGeocoder:(MKReverseGeocoder *)aGeocoder didFailWithError:(NSError *)error;
{
	[[NSNotificationCenter defaultCenter] postNotificationName:CWPointStatusDidChangeNotification object:[CWPoint currentLocationPoint]];
  [aGeocoder release];
  geocoder = nil;
}

- (void)reverseGeocoder:(MKReverseGeocoder *)aGeocoder didFindPlacemark:(MKPlacemark *)placemark;
{
  NSString* name = placemark.thoroughfare;
  if (placemark.subThoroughfare != nil) {
  	name = [name stringByAppendingFormat:@" %@", placemark.subThoroughfare];
  }
  if (placemark.locality != nil) {
	  name = [name stringByAppendingFormat:@", %@", placemark.locality];
  }
  if (![placemark.countryCode isEqualToString:@"SE"]) {
  	name = [name stringByAppendingFormat:@", %@", placemark.country];
  }
	[[CWPoint currentLocationPoint] setValue:name forKey:@"name"];
	[[NSNotificationCenter defaultCenter] postNotificationName:CWPointStatusDidChangeNotification object:[CWPoint currentLocationPoint]];
  [aGeocoder release];
  geocoder = nil;
}

-(void)cancelLocationUpdates;
{
  [locationManager stopUpdatingLocation];
  isUpdatingLocation = NO;
	[[NSNotificationCenter defaultCenter] postNotificationName:CWPointStatusDidChangeNotification object:[CWPoint currentLocationPoint]];
}

- (void)locationManager:(CLLocationManager *)manager didUpdateToLocation:(CLLocation *)newLocation fromLocation:(CLLocation *)oldLocation;
{
  [[NSRunLoop currentRunLoop] cancelPerformSelector:@selector(cancelLocationUpdates) target:self argument:nil];
	[lastLocation release];
  lastLocation = [newLocation retain];
  CLLocationAccuracy accuracy = MAX(lastLocation.horizontalAccuracy, lastLocation.verticalAccuracy);
  if (accuracy < CWPointAccuracyThreshold) {
  	[manager stopUpdatingLocation];
    isUpdatingLocation = NO;
  } else {
		[self performSelector:@selector(cancelLocationUpdates) withObject:nil afterDelay:CWPointMaxTimeToWaitForLocation];
  }
  if (geocoder != nil) {
  	[geocoder cancel];
    [geocoder release];
  }
  geocoder = [[MKReverseGeocoder alloc] initWithCoordinate:newLocation.coordinate];
  geocoder.delegate = self;
  [geocoder start];
}

-(void)didSelectPoint;
{
	if (self.ID == CWPointCurrentLocationID) {
    failedLocation = NO;
    isUpdatingLocation = YES;
  	[locationManager startUpdatingLocation];
  }
}

-(NSString*)queryString;
{
  if (self.ID == CWPointCurrentLocationID) {
    CLLocationCoordinate2D coordinate = lastLocation.coordinate;
    //coordinate.latitude = 55.612150;
    //coordinate.longitude = 12.995901;
	  return [NSString stringWithFormat:@"%%7C%f%%3A%f%%7C3", coordinate.latitude, coordinate.longitude];
  } else {
	  return [NSString stringWithFormat:@"%%7C%d%%7C%d", self.ID, CWQueryTypeForPointType(self.type)];
  }
}

-(CLLocation*)location;
{
	if (self.ID == CWPointCurrentLocationID) {
		return lastLocation;
  } else {
  	return nil;
  }
}

-(CLLocationAccuracy)locationAccuracy;
{
	if (lastLocation != nil) {
    return MAX(lastLocation.horizontalAccuracy, lastLocation.verticalAccuracy);
  } else {
  	return 10000;
  }
}

-(BOOL)isReady;
{
	return self.pointStatus <= CWPointStatusInitial;
}

-(CWPointStatus)pointStatus;
{
	if (self.ID == CWPointCurrentLocationID) {
		if (failedLocation) {
    	return CWPointStatusError;
    } else {
			return lastLocation == nil ? CWPointStatusPending : (isUpdatingLocation ? CWPointStatusInitial : CWPointStatusKnown);
    }
  } else {
  	return CWPointStatusStatic;
  }
}

+(CWPoint*)pointWithID:(NSUInteger)pointId name:(NSString*)name type:(CWPointType)type;
{
	CWPoint* point = [[CWPoint alloc] init];
  if (point) {
		point->_ID = pointId;
    point->_name = [name retain];
    point->_type = type;
    if (pointId == CWPointCurrentLocationID) {
      locationManager = [[CLLocationManager alloc] init];
      locationManager.desiredAccuracy = kCLLocationAccuracyNearestTenMeters;
      locationManager.distanceFilter = kCLDistanceFilterNone;
      locationManager.delegate = point;
    }
  }
  return [point autorelease];
}

+(CWPoint*)currentLocationPoint;
{
	static CWPoint* currentLocationPoint = nil;
  if (currentLocationPoint == nil) {
		currentLocationPoint = [[CWPoint pointWithID:CWPointCurrentLocationID
                                            name:NSLocalizedString(@"CurrentLocation", nil)
                                            type:CWPointTypeGPSCoordinte] retain];
  }
  return currentLocationPoint;
}

+(NSMutableArray*)defaultKnownPoints;
{
  NSMutableArray* points = [NSMutableArray array];
	[points addObject:[CWPoint pointWithID:80000 name:@"Malmö C" type:CWPointTypeStopArea]];
	[points addObject:[CWPoint pointWithID:80120 name:@"Södervärn Malmö" type:CWPointTypeStopArea]];
	[points addObject:[CWPoint pointWithID:81216 name:@"Lund C" type:CWPointTypeStopArea]];
	[points addObject:[CWPoint pointWithID:83241 name:@"Helsingborg C" type:CWPointTypeStopArea]];
	[points addObject:[CWPoint pointWithID:85131 name:@"Eslövs Station" type:CWPointTypeStopArea]];
  return points;
}

+(NSMutableArray*)defaultJourneySearchBookmarks;
{
  return [NSMutableArray array];
}

+(NSArray*)pointsMatchingString:(NSString*)searchString;
{
  CWPointParser* pointParser = [[CWPointParser alloc] initWithSearchString:searchString];
  NSArray* points = [pointParser parsePoints];
  NSLog(@"Points: %@", [points description]);
	[pointParser release];
  return points;
}

-(id)initWithCoder:(NSCoder *)aDecoder;
{
  NSUInteger pointId = [aDecoder decodeIntegerForKey:@"ID"];
 	if (pointId == CWPointCurrentLocationID) {
		[self release];
    return [CWPoint currentLocationPoint];
  } else {
    self = [self init];
    if (self) {
      _ID = pointId;
      _name = [[aDecoder decodeObjectForKey:@"name"] retain];
      _stopPoint = [[aDecoder decodeObjectForKey:@"stopPoint"] retain];
      _type = [aDecoder decodeIntegerForKey:@"type"];
    }
    return self;
  }
}

-(void)encodeWithCoder:(NSCoder *)aCoder;
{
  [aCoder encodeInteger:self.ID forKey:@"ID"];
  [aCoder encodeObject:self.name forKey:@"name"];
  [aCoder encodeObject:self.stopPoint forKey:@"stopPoint"];
  [aCoder encodeInteger:self.type forKey:@"type"];
}

-(void)dealloc;
{
  [_name release];
  [super dealloc];
}

-(UIImage*)imageForType;
{
  if (self.ID == CWPointCurrentLocationID) {
 		return [UIImage imageNamed:@"point-current.png"];
  } else {
	  return [UIImage imageNamed:[NSString stringWithFormat:@"point-%d.png", self.type]];
  }
}

-(UITableViewCellAccessoryType)accessoryType;
{
  return UITableViewCellAccessoryNone;
  //	return self.type == CWPointTypeStopArea ? UITableViewCellAccessoryDetailDisclosureButton : UITableViewCellAccessoryNone;
}

-(BOOL)isEqual:(id)anObject;
{
  if (self == anObject) {
    return YES;
  } else if ([anObject isKindOfClass:[CWPoint class]]) {
    CWPoint* otherPoint = anObject;
    // It is safe to ignore types, as they use different ID ranges anyway.
    return self.ID == otherPoint.ID;
  }
  return NO;
}

-(NSString*)description;
{
  return [NSString stringWithFormat:@"<CWPoint: %d, '%@', %@>", self.ID, self.name, self.typeString];
}

@end

@implementation CWPointParser

+(NSURL*)queryURLForSearchString:(NSString*)searchString;
{
  NSString* escapedSearchString = [searchString stringByAddingPercentEscapesUsingEncoding:NSUTF8StringEncoding];
  NSString* query = [NSString stringWithFormat:@"querypage.asp?inppointto=zq&inppointfr=%@", escapedSearchString];
  query = [[[[NSBundle mainBundle] infoDictionary] objectForKey:@"ServerURL"] stringByAppendingString:query];
  return [NSURL URLWithString:query];
}

-(id)initWithSearchString:(NSString*)searchString;
{
  self = [super init];
  if (self) {
    url = [[CWPointParser queryURLForSearchString:searchString] retain];
  }
  return self;
}

-(void)dealloc;
{
  [url release];
  [super dealloc];
}

-(NSArray*)parsePoints;
{
  [UIApplication sharedApplication].networkActivityIndicatorVisible = YES;
  NSDictionary* translation = [NSDictionary dictionaryWithContentsOfFile:[[NSBundle mainBundle] pathForResource:@"CWPointParser" ofType:@"plist"]];
	CWXMLTranslater* translater = [[CWXMLTranslater alloc] initWithTranslationPropertyList:translation delegate:self];
	NSError* error = nil;
  NSArray* points = [translater translateContentsOfURL:url delegate:translater.delegate error:&error];
  if (points == nil) {
    NSLog(@"Error: %@", [error description]);
    NSLog(@"URL: %@", [url description]);
    NSLog(@"XML: %@", [NSString stringWithContentsOfURL:url encoding:NSUTF8StringEncoding error:NULL]);
    // Intentionally not an error, because this happens too often!
    /*
  	UIAlertView* alert = [[UIAlertView alloc] initWithTitle:NSLocalizedString(@"FailedNetwork", nil) message:NSLocalizedString(@"FailedNetworkMessage", nil) delegate:nil cancelButtonTitle:NSLocalizedString(@"OK", nil) otherButtonTitles:nil];
    [alert show];
    [alert release];
     */
  }
  [UIApplication sharedApplication].networkActivityIndicatorVisible = NO;
	return points;
}

-(id)xmlTranslater:(CWXMLTranslater*)translater willTranslateValueOfClass:(Class)aClass withString:(NSString*)aString forKey:(NSString*)key;
{
	if ([key isEqualToString:@"type"]) {
    CWPointType type;
    if ([aString isEqualToString:@"STOP_AREA"]) {
      type = CWPointTypeStopArea;
    } else if ([aString isEqualToString:@"ADDRESS"]) {
      type = CWPointTypeAddress;
    } else if ([aString isEqualToString:@"POI"]) {
      type = CWPointTypePointOfInterest;
    }
    return [NSNumber numberWithInt:(int)type];
  } else if ([key isEqualToString:@"name"]) {
    return [aString capitalizedString];
  }
  return nil;
}

@end
